#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os.path as p
import typing as t
import watchdog.observers
import watchdog.events


class Watcher:
    def __init__(self, file: str):
        if not p.isfile(file):
            raise FileNotFoundError(file)
        self._fp = file
        self._handlers = []
        self._observer = watchdog.observers.Observer()
        self._observer.schedule(self, file)

    def __enter__(self) -> 'Watcher':
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @property
    def file(self) -> str:
        return self._fp

    def on_change(self, handler: t.Callable[[], None]):
        self._handlers.append(handler)
        return handler

    def start(self) -> 'Watcher':
        self._observer.start()
        return self

    def stop(self):
        self._observer.stop()
        self._observer.join()

    def dispatch(self, event: watchdog.events.FileSystemEvent):
        if event.event_type != watchdog.events.EVENT_TYPE_MODIFIED:
            return
        if event.src_path != self.file:
            return
        for handler in self._handlers:
            handler()  # maybe add some sort of catch or so later



# -*- coding=utf-8 -*-
r"""

"""
import typing as t
if t.TYPE_CHECKING:
    from pathlib import Path
from .util import KEY, INDEX, Convert, unify_key


__all__ = ['ConfigInterface']


MISSING = object()
_T = t.TypeVar('_T')


class ConfigInterface:
    r"""
    interface to a deep object

    todo: allow key-variants
    config.get('web', ['host', 'bind])
    """
    _obj: t.Dict[str, t.Any]

    def __init__(self, obj: t.Union['ConfigInterface', t.Dict[str, t.Any]] = None):
        self._obj = obj._obj if isinstance(obj, ConfigInterface) else (obj or {})

    # ---------------------------------------------------------------------------------------------------------------- #

    def __str__(self):
        return f"<{', '.join(self)}>"

    def __repr__(self):
        return f"<{type(self).__name__}: {', '.join(self)}>"

    def __format__(self, format_spec):
        import re
        raw_keys = re.split(r'\s*[.,:;|]\s*', format_spec, flags=re.I)
        keys = [int(k) if k.isdecimal() else k for k in raw_keys if k]
        return self.getstr(*keys, fallback="<missing>")

    # ---------------------------------------------------------------------------------------------------------------- #

    def __len__(self):
        return len(self._obj)

    def __iter__(self):
        yield from self._obj

    # ---------------------------------------------------------------------------------------------------------------- #

    def __getitem__(self, item: KEY):
        obj = self._obj
        for key in unify_key(item):
            obj = obj[key]
        return obj

    def get(self, *keys: INDEX, fallback: t.Any = MISSING, converter: t.Callable[..., t.Any] = None, **kwargs):
        r"""get the configuration value"""
        if converter is None and kwargs:
            import warnings
            warnings.warn(f"converter arguments without converter passed", SyntaxWarning, stacklevel=2)
        try:
            value = self[keys]
            return converter(value, **kwargs) if converter is not None else value
        except (KeyError, IndexError, TypeError) as exc:
            if fallback is not MISSING:
                return fallback
            raise exc

    def getstr(self, *keys: INDEX, fallback: t.Any = MISSING) -> str:
        r"""Ensures the returned type is str"""
        return self.get(*keys, fallback=fallback, converter=Convert.to_str)

    def getint(self, *keys: INDEX, fallback: t.Any = MISSING) -> int:
        r"""Ensures the returned type is int"""
        return self.get(*keys, fallback=fallback, converter=Convert.to_int)

    def getfloat(self, *keys: INDEX, fallback: t.Any = MISSING) -> float:
        r"""Ensures the returned type is float"""
        return self.get(*keys, fallback=fallback, converter=Convert.to_float)

    def getbool(self, *keys: INDEX, fallback: t.Any = MISSING) -> bool:
        r"""Ensures the returned type is bool"""
        return self.get(*keys, fallback=fallback, converter=Convert.to_bool)

    def getboolean(self, *args, **kwargs) -> bool:
        r"""Ensures the returned type is bool"""
        import warnings
        warnings.warn("getboolean is deprecated. Use .getbool() instead", DeprecationWarning, stacklevel=2)
        return self.getbool(*args, **kwargs)

    def getlist(self, *keys: INDEX, fallback: t.Any = MISSING, cast: t.Type[_T] = t.Any) -> t.List[_T]:
        r"""Ensures the returned type is list. (you may prefer .getsplit())"""
        return self.get(*keys, fallback=fallback, converter=Convert.to_list, cast=cast)

    def gettuple(self, *keys: INDEX, fallback: t.Any = MISSING, cast: t.Type[_T] = t.Any) -> t.Tuple[_T, ...]:
        r"""Ensures the returned type is tuple. (you may prefer `tuple(.getsplit())`)"""
        return self.get(*keys, fallback=fallback, converter=Convert.to_tuple, cast=cast)

    def getsplit(self, *keys: INDEX, fallback: t.Any = MISSING, cast: t.Type[_T] = str) -> t.List[_T]:
        r"""split (and trim) by , or ;"""
        return self.get(*keys, fallback=fallback, converter=Convert.split, cast=cast)

    def getshlex(self, *keys: INDEX, fallback: t.Any = MISSING) -> t.List[str]:
        r"""split like the command line"""
        return self.get(*keys, fallback=fallback, converter=Convert.split_shlex)

    def getpath(self, *keys: INDEX, fallback: t.Any = MISSING) -> 'Path':
        r"""get as pathlib.Path"""
        return self.get(*keys, fallback=fallback, converter=Convert.to_path)

    @t.overload
    def getpaths(self, *keys: INDEX, fallback: t.Any = MISSING, as_path: t.Literal[True] = ...) -> t.List['Path']: ...
    @t.overload
    def getpaths(self, *keys: INDEX, fallback: t.Any = MISSING, as_path: t.Literal[False]) -> t.List[str]: ...

    def getpaths(self, *keys: INDEX, fallback: t.Any = MISSING, as_path: bool = True) -> t.List[t.Union[str, 'Path']]:
        r"""split by os.path.pathsep (returns: pathlib.Path if as_path else str)"""
        return self.get(*keys, fallback=fallback, converter=Convert.split_paths, as_path=as_path)

    def getinterface(self, *keys: INDEX, fallback: dict = MISSING) -> 'ConfigInterface':
        r"""returns a new ConfigInterface of given option"""
        val = self.get(*keys, fallback=MISSING if fallback is MISSING else fallback)
        if not isinstance(val, dict):
            raise TypeError(val)
        return type(self)(val)

    def gettype(self, *keys: INDEX, fallback: t.Any = MISSING) -> type:
        r"""gets the class of given configuration-option"""
        return type(self.get(*keys, fallback=fallback))

    # ---------------------------------------------------------------------------------------------------------------- #

    def __delitem__(self, key: KEY):
        *keys, del_key = unify_key(key)
        obj = self._obj
        for key in keys:
            obj = obj[key]
        del obj[del_key]

    def remove(self, *keys: INDEX, failok: bool = False):
        r"""removes a given configuration"""
        try:
            del self[keys]
        except (KeyError, IndexError, TypeError) as exc:
            if not failok:
                raise exc

    # ---------------------------------------------------------------------------------------------------------------- #

    def __contains__(self, item: KEY):
        try:
            obj = self._obj
            for key in unify_key(item):
                obj = obj[key]
        except (KeyError, IndexError, TypeError):
            return False
        else:
            return True

    def has(self, *keys: INDEX) -> bool:
        r"""checks if configuration exists"""
        return keys in self

    # ---------------------------------------------------------------------------------------------------------------- #

    def update(self, __other: t.Union['ConfigInterface', dict] = None, **kwargs):
        r"""update the configuration with replacing"""
        if isinstance(__other, ConfigInterface):
            __other = __other._obj
        other = (__other | kwargs) if __other else kwargs

        self._obj.update(other)

    def merge(self, __other: t.Union['ConfigInterface', dict] = None, **kwargs):
        r"""update the configuration with merging"""
        if isinstance(__other, ConfigInterface):
            __other = __other._obj
        other = (__other | kwargs) if __other else kwargs

        def merge_dict(d: dict, o: dict):
            for key, value in o.items():
                if key in d and isinstance(d[key], dict):
                    merge_dict(d[key], value)
                else:
                    d[key] = value

        merge_dict(self._obj, other)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
from configlib.exceptions import NotSupportedError
from .conf_loader import load_conf
from .xml_loader import load_xml
from .json_loader import load_json
from .jsonc_loader import load_jsonc
try:
    from .json5_loader import load_json5
except NotSupportedError:
    def load_json5(*_, **__):
        raise NotSupportedError()
try:
    from .toml_loader import load_toml
except NotSupportedError:
    def load_toml(*_, **__):
        raise NotSupportedError()
try:
    from .yaml_loader import load_yaml
except NotSupportedError:
    def load_yaml(*_, **__):
        raise NotSupportedError()



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import typing as t
from . import places

T_PLACE: t.TypeAlias = t.Union[str, t.Callable[[], str]]
T_PLACES: t.TypeAlias = t.List[T_PLACE]

DEFAULT_PLACES: T_PLACES = [
    places.cwd,
    places.local,
    places.repository,
    places.localappdata,
    places.user_conf,
    places.home,
    places.etc,
]



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
from pathlib import Path
from ..exceptions import ConfigNotFoundError
from ._variations import variations as iter_variations
from .default_places import DEFAULT_PLACES, T_PLACES


def find(*variants: str, places: T_PLACES = None) -> Path:
    r"""
    find a config file

    note: if a variant ends with .ext then every supported config-type is searched for

    note: you can make variations with file.{ext1,ext2}

    note: absolute paths have higher priority than relative

    :param variants: name-variants of the config file
    :param places: list of directories to search in
    :return: pathlib.Path
    :raise ConfigNotFoundError: config-file could not be found
    """
    if places is None:
        places = DEFAULT_PLACES

    variations = list(iter_variations(*variants))

    # split by absolute or relative
    absolutes, relatives = [], []

    for variation in map(Path, variations):
        if variation.is_absolute():
            absolutes.append(variation)
        else:
            relatives.append(variation)

    # check absolutes
    for fp in absolutes:
        if fp.is_file():
            return fp

    # check relatives
    for place in places:
        if callable(place):
            place = place()
        if place is None:  # not-supported on this platform
            continue

        place = Path(place)
        if not place.is_dir():  # doesn't exist. No need to continue searching
            continue

        for variation in relatives:
            fp = place / variation
            if fp.is_file():
                return fp

    raise ConfigNotFoundError("no config-file could be found anywhere")



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import textwrap
import typing as t
from xml.etree import ElementTree
from configlib.loader import register_loader


@register_loader('xml')
def load_xml(fp: t.Union[str, os.PathLike]) -> dict:
    r"""
    <?xml version="1.0" encoding="UTF-8"?>
    <message>
        <warning>
             Hello World
        </warning>
    </message>
    """
    with open(fp, 'r') as file:
        root = ElementTree.parse(file).getroot()
    return {root.tag: node2dict(root)}


def node2dict(node: ElementTree.Element) -> dict:
    obj: t.Dict[str, t.Union[str, t.Dict, t.List[t.Union[str, t.List, t.Dict]]]] = {
        f'@{attr}': value
        for attr, value in node.attrib.items()
    }
    for child in node:
        if child.tag in obj:
            if not isinstance(obj[child.tag], list):
                obj[child.tag] = [obj[child.tag]]
            obj[child.tag].append(node2dict(child))
        else:
            obj[child.tag] = node2dict(child)
    text = textwrap.dedent(node.text).strip()
    if text:
        obj['#'] = text
    return obj



# -*- coding=utf-8 -*-
r"""

"""
import typing as t
from ..interface.typing import INDEX


XMISSING = object()
T = t.TypeVar("T")


def get_any(getter: t.Callable[..., T], *variants: t.Iterable[INDEX], fallback: t.Any = XMISSING) -> T:
    r"""
    config = ConfigInterface()
    get_any(config.getstr, ['logging', 'console', 'level'], ['logging', 'level'], fallback="INFO")

    :param getter: ConfigInterface.get*() method
    :param variants: access variants
    :param fallback: fallback value
    :return:
    """
    for keys in variants:
        value = getter(*keys, fallback=XMISSING)
        if value is not XMISSING:
            return value
    if fallback is XMISSING:
        raise KeyError(f"can't get any for {getter.__name__}")
    else:
        return fallback



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import json
import typing as t
from configlib.loader import register_loader


@register_loader('json')
def load_json(fp: t.Union[str, os.PathLike]) -> dict:
    r"""
    {
      "first_name": "John",
      "last_name": "Smith",
      "is_alive": true,
      "age": 27,
      "address": {
        "street_address": "21 2nd Street",
        "city": "New York",
        "state": "NY",
        "postal_code": "10021-3100"
      },
      "phone_numbers": [
        {
          "type": "home",
          "number": "212 555-1234"
        },
        {
          "type": "office",
          "number": "646 555-4567"
        }
      ],
      "children": [
        "Catherine",
        "Thomas",
        "Trevor"
      ],
      "spouse": null
    }
    """
    with open(fp, 'r') as file:
        return json.load(file)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import os.path as p


def fsroot():
    r"""
    should return 'C:\' under Windows and '/' under good OS's
    """
    return p.abspath(p.sep)


def cwd():
    r"""
    current working directory
    """
    return p.abspath(os.getcwd())


def local():
    r"""
    root-directory of the python script currently executed
    """
    import sys
    main = sys.modules.get('__main__')
    if main is None or not hasattr(main, '__file__'):
        return None
    return p.abspath(
        p.dirname(main.__file__)
    )


def repository():
    r"""
    folder that contains a .git directory somewhere above local()
    """
    directory = local()
    if directory is None:
        return None

    while not p.isdir(p.join(directory, '.git')) and directory != '/':
        directory = p.dirname(directory)

    if directory == '/':
        return None

    return directory


def user_conf():
    r"""
    the .config directory in the home-directory of the current user
    ~/.config/
    """
    return p.join(
        home(), ".config"
    )


def home():
    r"""
    the home-directory of the current user
    ~/
    """
    return p.abspath(p.expanduser("~"))


def etc():
    r"""
    the etc folder in the filesystem root directory
    /etc/
    """
    return p.join(fsroot(), "etc")


def localappdata():
    r"""
    %LOCALAPPDATA%
    """
    return os.getenv("LOCALAPPDATA")



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import unittest
from pathlib import Path


HERE = Path(__file__).parent


class TestLoaders(unittest.TestCase):
    def test_conf(self):
        from configlib import load

        config = load(HERE / "test_configs" / "app.conf")
        self.assertEqual(config.getstr('owner', 'name'), "John Doe")
        self.assertEqual(config.getint('database', 'port'), 143)

    def test_json(self):
        from configlib import load

        config = load(HERE / "test_configs" / "app.json")
        self.assertEqual(config.getstr('first_name'), "John")
        self.assertEqual(config.getstr('phone_numbers', 0, 'type'), "home")

    def test_jsonc(self):
        from configlib import load

        config = load(HERE / "test_configs" / "app.jsonc")
        self.assertFalse(config.getbool('true'))
        self.assertEqual(config.getint('number'), 42)

    def test_json5(self):
        from configlib import load

        config = load(HERE / "test_configs" / "app.json5")
        self.assertEqual(config.get('positiveSign'), 1)

    def test_toml(self):
        from configlib import load

        config = load(HERE / "test_configs" / "app.toml")
        self.assertEqual(config.getstr('title'), "TOML Example")
        self.assertTrue(config.getbool('database', 'enabled'))

    def test_xml(self):
        from configlib import load

        config = load(HERE / "test_configs" / "app.xml")
        self.assertEqual(config.getstr('message', 'warning', '#'), "Hello World")

    def test_yaml(self):
        from configlib import load

        config = load(HERE / "test_configs" / "app.yaml")
        self.assertEqual(config.getstr('date'), "2012-08-06")
        self.assertEqual(config.getint('items', 1, 'quantity'), 1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import typing as t
import configparser
from configlib.loader import register_loader


@register_loader('ini', 'conf', 'config')
def load_conf(fp: t.Union[str, os.PathLike]) -> dict:
    r"""
    # last modified 1 April 2001 by John Doe
    [owner]
    name = John Doe
    organization = Acme Widgets Inc.

    [database]
    # use IP address in case network name resolution is not working
    server = 192.0.2.62
    port = 143
    file = "payroll.dat"
    """
    parser = configparser.ConfigParser(
        allow_no_value=False,
        strict=True,
        empty_lines_in_values=False,
        interpolation=configparser.ExtendedInterpolation(),
    )
    parser.optionxform = lambda option: option.lower().replace('-', '_')  # 'Hello-World' => 'hello_world'
    with open(fp, 'r') as file:
        parser.read_file(file)
    return {
        section: {
            option: parser.get(section, option)
            for option in parser.options(section)
        } for section in parser.sections()
    }



# -*- coding=utf-8 -*-
r"""

"""
import re
import itertools
import typing as t
from ..loader import get_supported_formats


PATTERN = re.compile(r"{(.+?)}")
WILDCARD_EXT = re.compile(r"\.ext$")
ALL_EXTENSIONS = f".{{{','.join(get_supported_formats())}}}"


def variations(*variants: str) -> t.Iterator[str]:
    r"""
    variations("hello.{yaml,yml}") -> ["hello.yaml", "hello.yml"]
    variations("hello.ext") -> ["hello.ini", "hello.conf", "hello.json", ...]
    """
    for variant in variants:
        variant = WILDCARD_EXT.sub(ALL_EXTENSIONS, variant)  # "hello.ext" => "hello.{ini,conf,json,...}"
        patterns: t.List[str] = PATTERN.findall(variant)
        if not patterns:
            yield variant
            continue
        patterns: t.List[t.List[str]] = [pattern.split(",") for pattern in patterns]
        template = PATTERN.sub('{}', variant)
        for values in itertools.product(*patterns):
            yield template.format(*values)



# -*- coding=utf-8 -*-
r"""

"""
import re
import typing as t


def split_double(name: str) -> t.List[str]:
    return re.split(r"_{2,}", name)


def split_single(name: str) -> t.List[str]:
    return re.split(r"_+", name)


default_split = split_double


def transform_lower(name: str) -> str:
    return name.lower().replace('-', '_')


def transform_upper(name: str) -> str:
    return name.upper().replace('-', '_')


default_transform = transform_lower



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""


class ConfigLibError(Exception):
    r"""
    any exception coming from configlib
    """
    pass


class ConfigNotFoundError(ConfigLibError):
    r"""
    A configuration file could not be found
    """
    pass


class NotSupportedError(ConfigLibError):
    r"""
    a configuration file could not be parsed
    """
    pass


class CallOrderError(ConfigLibError):
    r"""
    another method should have been called first
    """
    pass



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import unittest


class TestImport(unittest.TestCase):
    def test_import(self):
        import configlib  # noqa


if __name__ == '__main__':
    unittest.main()
